"""
Module for database migration using Alembic
It checks if database is accesible.
Creates sqlite database if needed.
Checks if database have SqlAlchemy tables. If not - it creates tables.
    When tables are created, they are filled with default values.
    Such as AxFieldType are filled with field types.
Checks that current Database schema fits SqlAlchemy schema. If it is not,
    it tryes to upgrade it by running alembic migration scripts.

"""
import os
import sys
from loguru import logger
from alembic.config import Config
from alembic.migration import MigrationContext
from alembic.autogenerate import compare_metadata
from alembic import command
import backend.model as ax_model
import backend.misc as ax_misc
from backend.model import AxAlembicVersion, AxFieldType

this = sys.modules[__name__]
alembic_cfg = None


def init_alembic_config() -> None:
    """Create Alembic config"""
    try:
        ini_path = ax_misc.path("alembic.ini")
        this.alembic_cfg = Config(ini_path)
        alembic_folder = ax_misc.path("backend/alembic")
        this.alembic_cfg.set_main_option(
            'script_location', str(alembic_folder))
    except Exception:
        logger.exception('Error initating alembic config.')
        raise


def tables_exist() -> bool:
    """Checks if database tables are created"""
    try:
        tables_created = ax_model.engine.dialect.has_table(
            ax_model.engine, '_ax_alembic_version')
        if tables_created:
            return True
        else:
            return False
    except Exception:
        logger.exception('Error in detecting if database tables are created')
        raise


def create_tables() -> None:
    """Create database and create baseline version in Alembic"""
    with ax_model.scoped_session("migration -> create_tables") as db_session:
        if os.environ.get('AX_DB_REVISION') is None:
            msg = 'Cant find AX_DB_REVISION in enviroment variables or app.yaml'
            logger.error(msg)
            raise Exception(msg)

        ax_model.Base.query = db_session.query_property()
        ax_model.Base.metadata.create_all(ax_model.engine)
        first_version = AxAlembicVersion()
        first_version.version_num = os.environ.get('AX_DB_REVISION')
        db_session.add(first_version)

        logger.info('Ax tables not found. Creating database tables.')
        return True


def create_field_types() -> None:
    """Fill AxFieldType with field types"""
    try:
        with ax_model.scoped_session() as db_session:
            # Identification
            db_session.add(AxFieldType(
                tag='group-id',
                is_group=True,
                position=1,
                icon="key"))
            db_session.add(AxFieldType(
                tag="AxGuid",
                parent="group-id",
                default_db_name="guid",
                position=1,
                is_virtual=True,
                comparator="",
                icon="key"))
            db_session.add(AxFieldType(
                tag='AxNum',
                parent="group-id",
                default_db_name="axNum",
                value_type="VARCHAR(255)",
                position=2,
                is_readonly=True,
                is_backend_available=True,
                icon="sort-numeric-up"))

            # Process controll
            db_session.add(AxFieldType(
                tag='group-process',
                is_group=True,
                position=2,
                icon="code-branch"))
            db_session.add(AxFieldType(
                tag="AxState",
                parent="group-process",
                default_db_name="axState",
                position=1,
                is_virtual=True,
                icon="code-branch"))
            db_session.add(AxFieldType(
                tag="AxChangelog",
                parent="group-process",
                position=4,
                default_db_name="changelog",
                is_columnn_avalible=True,
                value_type="JSON",
                is_backend_available=True,
                is_readonly=True,
                is_updated_always=True,
                is_always_whole_row=True,
                icon="code-branch"))

            # String
            db_session.add(AxFieldType(
                tag='group-string',
                is_group=True,
                position=3,
                icon="font"))
            db_session.add(AxFieldType(
                tag="AxString",
                parent="group-string",
                default_db_name="string",
                position=1,
                value_type="VARCHAR(255)",
                is_inline_editable=True,
                icon="font"))
            db_session.add(AxFieldType(
                tag="AxEmail",
                parent="group-string",
                default_db_name="email",
                position=2,
                value_type="VARCHAR(255)",
                is_inline_editable=True,
                icon="font"))
            db_session.add(AxFieldType(
                tag="AxTelephone",
                parent="group-string",
                default_db_name="telephone",
                position=3,
                value_type="VARCHAR(255)",
                is_columnn_avalible=True,
                is_inline_editable=True,
                icon="font"))

            # Text
            db_session.add(AxFieldType(
                tag='group-text',
                name="types.text",
                is_group=True,
                position=4,
                icon="align-left"))
            db_session.add(AxFieldType(
                tag="AxText",
                parent="group-text",
                default_db_name="text",
                position=1,
                value_type="TEXT",
                is_inline_editable=True,
                icon="align-left"))
            db_session.add(AxFieldType(
                tag="AxMarkdown",
                parent="group-text",
                default_db_name="markdown",
                position=1,
                value_type="TEXT",
                is_inline_editable=True,
                icon="align-left"))

            # Number
            db_session.add(AxFieldType(
                tag='group-number',
                is_group=True,
                position=5,
                icon="hashtag"))
            db_session.add(AxFieldType(
                tag="AxInteger",
                parent="group-number",
                default_db_name="integer",
                position=1,
                value_type="INT",
                comparator="number",
                is_inline_editable=True,
                is_columnn_avalible=True,
                icon="hashtag"))
            db_session.add(AxFieldType(
                tag="AxDecimal",
                parent="group-number",
                default_db_name="decimal",
                position=2,
                value_type="DECIMAL(65,2)",
                comparator="number",
                is_columnn_avalible=True,
                is_inline_editable=True,
                icon="hashtag"))
            db_session.add(AxFieldType(
                tag="AxIntSlider",
                parent="group-number",
                default_db_name="intSlider",
                position=1,
                value_type="INT",
                comparator="number",
                is_inline_editable=True,
                icon="hashtag"))

            # Boolean
            db_session.add(AxFieldType(
                tag='group-boolean',
                is_group=True,
                position=6,
                icon="toggle-on"))
            db_session.add(AxFieldType(
                tag="AxCheckbox",
                parent="group-boolean",
                default_db_name="checkbox",
                position=1,
                value_type="BOOL",
                is_inline_editable=True,
                is_columnn_avalible=True,
                icon="toggle-on"))

            # Relationship
            db_session.add(AxFieldType(
                tag='group-relationship',
                is_group=True,
                position=7,
                icon="link"))
            db_session.add(AxFieldType(
                tag="Ax1to1",
                parent="group-relationship",
                default_name="types.ax-1to1-default",
                default_db_name="toOne",
                position=1,
                value_type="GUID",
                is_inline_editable=False,
                is_columnn_avalible=True,
                icon="link"))
            db_session.add(AxFieldType(
                tag="Ax1tom",
                parent="group-relationship",
                default_name="types.ax-1tom-default",
                default_db_name="toMany",
                position=2,
                value_type="JSON",
                is_backend_available=True,
                is_columnn_avalible=True,
                is_inline_editable=False,
                icon="link"))
            db_session.add(AxFieldType(
                tag="Ax1tomTable",
                parent="group-relationship",
                default_name="types.ax-1tom-table-default",
                default_db_name="toManyTable",
                position=3,
                value_type="JSON",
                is_backend_available=True,
                is_columnn_avalible=True,
                is_inline_editable=True,
                icon="link"))

            #Date and Time
            db_session.add(AxFieldType(
                tag='group-date',
                is_group=True,
                position=8,
                icon="calendar"))
            db_session.add(AxFieldType(
                tag="AxDate",
                parent="group-date",
                default_db_name="date",
                position=1,
                value_type="TIMESTAMP",
                comparator="date",
                is_inline_editable=True,
                icon="calendar"))

            # List
            db_session.add(AxFieldType(
                tag='group-list',
                is_group=True,
                position=9,
                icon="list"))
            db_session.add(AxFieldType(
                tag="AxChoise",
                parent="group-list",
                default_db_name="choise",
                position=1,
                value_type="VARCHAR(255)",
                is_inline_editable=True,
                icon="list"))
            db_session.add(AxFieldType(
                tag="AxRadio",
                parent="group-list",
                default_db_name="radio",
                position=1,
                value_type="VARCHAR(255)",
                is_inline_editable=True,
                icon="list"))

            # Images
            db_session.add(AxFieldType(
                tag='group-images',
                is_group=True,
                position=10,
                icon="image"))
            db_session.add(AxFieldType(
                tag="AxImageCropDb",
                parent="group-images",
                position=4,
                default_db_name="cropImage",
                is_backend_available=True,
                value_type="VARCHAR(255)",
                icon="image"))

            # Files
            db_session.add(AxFieldType(
                tag='group-files',
                is_group=True,
                position=11,
                icon="file"))
            db_session.add(AxFieldType(
                tag="AxFiles",
                parent="group-files",
                position=1,
                default_db_name="files",
                value_type="JSON",
                is_backend_available=True,
                icon="file"))

            # Users
            db_session.add(AxFieldType(
                tag='group-users',
                is_group=True,
                position=12,
                icon="user"))
            db_session.add(AxFieldType(
                tag="AxUsers",
                parent="group-users",
                position=1,
                default_db_name="users",
                is_backend_available=True,
                value_type="JSON",
                icon="user"))

            # Communication
            db_session.add(AxFieldType(
                tag='group-communication',
                is_group=True,
                position=13,
                icon="comments"))
            db_session.add(AxFieldType(
                tag="AxComments",
                parent="group-communication",
                position=4,
                default_db_name="comments",
                value_type="JSON",
                is_always_whole_row=True,
                is_backend_available=True,
                icon="comments"))
    except Exception:
        logger.exception('Failed creating default AxFieldTypes')
        raise


def database_fits_metadata() -> None:
    """Compare metada and database"""
    try:
        with ax_model.engine.connect() as active_connection:
            context = MigrationContext.configure(active_connection)
            db_session = ax_model.Session()
            ax_model.Base.query = db_session.query_property()
            metadata = ax_model.Base.metadata
            diff = compare_metadata(context, metadata)
            return diff == []
    except Exception:
        logger.exception('Error in comparing metadata and database.')
        raise


def upgrade_database():
    """Run Alembic migration script and update db version"""
    try:
        logger.info('Database does not fit metadata. Upgrading database.')
        command.upgrade(this.alembic_cfg, "head")
        command.stamp(this.alembic_cfg, "head")
    except Exception:
        logger.exception('Error upgrading database with Alembic')
        raise


def init_migration():
    """Initiate migration module"""
    init_alembic_config()
    if tables_exist() is False:
        create_tables()
        create_field_types()
    else:
        pass
        # if database_fits_metadata() is False:
        #     upgrade_database()



"""Axy Class Model
Contains class structure of Ax storage.
"""

import os
import sys
from pathlib import Path
from contextlib import contextmanager
import uuid
from loguru import logger
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship
from sqlalchemy import Text, String, CHAR, Float, Unicode, Boolean, Integer
from sqlalchemy import TypeDecorator, Column, LargeBinary
from sqlalchemy import ForeignKey
from sqlalchemy.dialects.postgresql import UUID
# from sqlalchemy.pool import SingletonThreadPool

import ujson as json
import backend.misc as ax_misc

# TODO: replace String(2000) for json fields with dialect agnostic JSON field

this = sys.modules[__name__]
engine = None
# db_session = None
db_url = None
Base = declarative_base()
Session = None # pylint: disable=invalid-name

@contextmanager
def scoped_session(error_msg=None):
    """ Using the scoped_session contextmanager is
     best practice to ensure the session gets closed
     and reduces noise in code by not having to manually
     commit or rollback the db if a exception occurs.
     Example - with ax_model.scoped_session(msg) as db_session:
     """
    session = this.Session()
    try:
        yield session
        session.commit()
    except:
        if error_msg:
            logger.exception(f'⛏⁉️ - {error_msg}')
        session.rollback()
        raise
    finally:
        session.close()


@contextmanager
def try_catch(db_session, error_msg=None, no_commit=False):
    """ Catches database errors and rollsback session
    with ax_model.try_catch(info.context['session'], err) as db_session: """
    try:
        yield db_session
        if not no_commit:
            db_session.commit()
    except:
        if error_msg:
            logger.exception(f'catch - {error_msg}')
        db_session.rollback()
        raise


def init_model(dialect: str, host: str, port: str, login: str, password: str,
               database: str, sqlite_filename: str, sqlite_absolute_path: str
               ) -> None:
    """Initiate database model

    Args:
        dialect (str): Supported dialects: sqlite | mysql
        host (str): Database host
        port (str): Database port
        login (str): Database user
        password (str): Database user password
        database (str): Database schema name
        sqlite_filename (str): Sqlite database file name. (ax_sqlite.db)
        sqlite_absolute_path (str): Path to Sqlite file. If file does not exist
            it will be created
    """
    try:
        if dialect == 'sqlite':
            if sqlite_absolute_path is None:
                db_path = ax_misc.path(sqlite_filename)
                this.db_url = 'sqlite:///' + str(db_path)
            else:
                db_path = str(Path(sqlite_absolute_path) / sqlite_filename)
                this.db_url = 'sqlite:///' + db_path
        elif dialect == 'postgre':
            this.db_url = (
                f'postgresql://{login}:{password}@{host}:{port}/{database}')
        else:
            msg = 'This database dialect is not supported'
            logger.error(msg)
            raise Exception(msg)

        logger.debug('DB url = {url}', url=this.db_url)
        this.engine = create_engine(
            this.db_url,
            convert_unicode=True,
            pool_recycle=3600,
            pool_pre_ping=True)

        # poolclass=SingletonThreadPool,

        this.Session = sessionmaker(bind=this.engine)

        # this.db_session = scoped_session(sessionmaker(autocommit=False,
        #                                               autoflush=False,
        #                                               bind=this.engine))

        # this.Base.query = db_session.query_property()
    except Exception:
        logger.exception('Error initating SqlAlchemy model')
        raise


class GUID(TypeDecorator):  # pylint: disable=W0223
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    """
    impl = CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value


class AxAlembicVersion(Base):
    """Stores current Database version for alembic migration"""
    __tablename__ = '_ax_alembic_version'
    version_num = Column(String(255), primary_key=True)


class AxMetric(Base):
    """Stores varius key/values.
    1 - counters for AxNum field"""
    __tablename__ = '_ax_metrics'
    guid = Column(GUID(), primary_key=True,
                  default=uuid.uuid4, unique=True, nullable=False)
    key = Column(String(255), unique=True, nullable=False)
    value = Column(String(255))


class AxSchedulerJob(Base):
    """Table and metadata for apscheduler."""
    __tablename__ = '_ax_scheduler_jobs'
    id = Column(Unicode(191, _warn_on_bytestring=False), primary_key=True)
    next_run_time = Column(Float(25), index=True)
    job_state = Column(LargeBinary, nullable=False)


class AxForm(Base):
    """Stores Ax-Forms"""
    __tablename__ = '_ax_forms'
    guid = Column(GUID(), primary_key=True,
                  default=uuid.uuid4, unique=True, nullable=False)
    name = Column(String(255))
    db_name = Column(String(255))   # Must be PascalCase. It is used to create
    # database tables and graphql schema
    parent = Column(GUID())  # Parent form. Used for visual tree in AdminHome
    position = Column(Integer())  # Position in tree
    fields = relationship(
        "AxField", order_by="AxField.position", cascade="all, delete-orphan")
    grids = relationship("AxGrid", order_by="AxGrid.position")
    states = relationship("AxState", cascade="all, delete-orphan")
    actions = relationship("AxAction", cascade="all, delete-orphan")
    roles = relationship("AxRole", cascade="all, delete-orphan")
    tom_label = Column(String(255))  # Used for displaying rows in relation
    # fields. {{name}} will result in
    # displaying name field as form chip
    icon = Column(String(255))  # font-awesome key
    is_folder = Column(Boolean, unique=False, default=False)  # Used in tree
    row_guid = None  # Stores current row guid.
    current_state_name = ""
    current_state_object = None
    from_state_name = ""
    from_state_object = None
    avalible_actions = None
    modal_guid = None   # Used in web-socket subscriptions to determin if
    # currrent form needs to notify on action
    permissions = relationship(
        "AxRoleFieldPermission", cascade="all, delete-orphan")

    @property
    def db_fields(self):
        """Only AxFields that are database columns"""
        db_fields = []
        for field in self.fields:
            if field.is_tab is False and field.field_type.is_virtual is False:
                db_fields.append(field)
        return db_fields

    @property
    def no_tab_fields(self):
        """Only AxFields that are database columns"""
        db_fields = []
        for field in self.fields:
            if field.is_tab is False:
                db_fields.append(field)
        return db_fields

    def get_row_data(self):
        """Set current AxForm to match specific row in database table"""
        return "GET ROW DATA"

    def get_field(self, field_db_name):
        """ returns AxFiled by given db_name of field"""
        for field in self.fields:
            if field.db_name == field_db_name:
                return field
        return None


class AxFieldType(Base):
    """List of avalible ax field types"""
    __tablename__ = '_ax_field_types'
    # Same as vue component. Like AxNum, AxString
    tag = Column(String(64), primary_key=True, unique=True)
    name = Column(String(255))
    parent = Column(String(64))  # Used in tree. Parent is always group.
    position = Column(Integer())  # Used in tree.
    default_name = Column(String(255))  # Comes from locale
    default_db_name = Column(String(255))
    # TEXT/INY/DECIMAL... Used by dialect.py
    value_type = Column(String(255))
    # to determin how to store current field
    comparator = Column(String(255))    # string/number Used in AxGrid to know
    # how to sort current fields columns

    icon = Column(String(255))  # font-awesome key
    is_group = Column(Boolean, unique=False,
                      default=False)  # Used in types tree
    # virtual meens that field is calculated, not stored in database
    # DB column is not created
    is_virtual = Column(Boolean, unique=False, default=False)
    # readonly meens that field is set by backend or action.
    # DB column is created but data does not come from form.
    is_readonly = Column(Boolean, unique=False, default=False)
    # True if field type have custom ag-grid column and have inline edit
    is_inline_editable = Column(Boolean, unique=False, default=False)
    # True if field have backend python code. Wich is executed before/after
    # insert/update/delete. Check backend/fields/ for more info
    is_backend_available = Column(Boolean, unique=False, default=False)
    # True if field type have custom ag-grid column
    is_columnn_avalible = Column(Boolean, unique=False, default=False)
    # True meens that this field is updated even if was not modified in form.
    # from/after backend code of field is executed on every action
    # Example - Changelog must be updated always, on every action
    is_updated_always = Column(Boolean, unique=False, default=False)
    # True meens that current field type always must be displayed as whole row
    is_always_whole_row = Column(Boolean, unique=False, default=False)

    def __init__(self,
                 tag, name=None,
                 position=0,
                 default_name=None,
                 default_db_name="",
                 value_type="",
                 parent="",
                 icon="",
                 comparator="",
                 is_virtual=False,
                 is_readonly=False,
                 is_inline_editable=False,
                 is_backend_available=False,
                 is_columnn_avalible=False,
                 is_updated_always=False,
                 is_group=False,
                 is_always_whole_row=False
                 ):
        self.tag = tag
        self.name = name
        self.position = position
        self.default_name = default_name
        self.default_db_name = default_db_name
        self.value_type = value_type
        self.parent = parent
        self.is_group = is_group
        self.comparator = comparator
        self.icon = icon
        self.is_virtual = is_virtual
        self.is_readonly = is_readonly
        self.is_inline_editable = is_inline_editable
        self.is_backend_available = is_backend_available
        self.is_columnn_avalible = is_columnn_avalible
        self.is_updated_always = is_updated_always
        self.is_always_whole_row = is_always_whole_row


class AxField(Base):
    """List of fields in each form"""
    __tablename__ = '_ax_fields'
    guid = Column(GUID(), primary_key=True,
                  default=uuid.uuid4, unique=True, nullable=False)
    form_guid = Column(GUID(), ForeignKey('_ax_forms.guid'))
    form = relationship("AxForm")
    name = Column(String(255))
    db_name = Column(String(255))   # Must be camelCase. It used to create
    # database table columns and GQL fields
    position = Column(Integer())  # tree position
    options_json = Column(String(2000))  # Stores JSON with params used by Vue
    # component. It is passed to UI
    # Stores JSON that is used only in python actions. AxAction code or fields
    # before/after methods
    private_options_json = Column(Text(convert_unicode=True))
    field_type_tag = Column(String(64), ForeignKey('_ax_field_types.tag'))
    field_type = relationship("AxFieldType")
    # True if current field is form tab
    is_tab = Column(Boolean, unique=False, default=False)
    is_required = Column(Boolean, unique=False, default=False)
    is_whole_row = Column(Boolean, unique=False, default=False)
    parent = Column(GUID())
    value = None
    is_readonly = False
    needs_sql_update = False  # Flag that current field must be updated in DB

    @property
    def is_virtual(self):
        """Is current fieldtype is Virtual (calculated field)"""
        return self.field_type.is_virtual

    @property
    def options(self):
        """ return parsed options_json """
        return json.loads(self.options_json)

    @property
    def files_path(self):
        """ returns path to directory where files for
        current field would be stored """
        files_path = os.path.join(
            ax_misc.uploads_root_dir,
            'form_row_field_file',
            str(self.form.guid),
            str(uuid.UUID(str(self.form.row_guid))),
            str(str(self.guid))
        )
        return files_path


class AxGrid(Base):
    """Stores Ax grids"""
    __tablename__ = '_ax_grids'
    guid = Column(GUID(), primary_key=True,
                  default=uuid.uuid4, unique=True, nullable=False)
    name = Column(String(255))
    # Must be PascalCase, used in GQL schema Types
    db_name = Column(String(255))
    position = Column(Integer())    # Position in tree
    options_json = Column(String(2000))  # JSON key/value see
    # TheConstructorGridsDrawerSecond.vue
    # Columns widths stored here too
    form_guid = Column(GUID(), ForeignKey('_ax_forms.guid'))
    form = relationship("AxForm")
    is_default_view = Column(Boolean)
    columns = relationship(
        "AxColumn", order_by="AxColumn.position", cascade="all, delete-orphan")


class AxColumn(Base):
    """Stores columns for each grid"""
    __tablename__ = '_ax_columns'
    guid = Column(GUID(), primary_key=True,
                  default=uuid.uuid4, unique=True, nullable=False)
    position = Column(Integer())    # Position in grid and in tree
    options_json = Column(String(2000))  # Currently not used
    field_guid = Column(GUID(), ForeignKey('_ax_fields.guid'))
    field = relationship("AxField")
    grid_guid = Column(GUID(), ForeignKey('_ax_grids.guid'))
    grid = relationship("AxGrid")
    column_type = Column(String(50))  # Not used. Was used for responsive grid
    aggregation_type = Column(String(50), nullable=True) # Not used yet.


class AxGroup2Users(Base):
    """Stores info on groups and users relation. What users are in group X."""
    __tablename__ = '_ax_group2user'
    guid = Column(GUID(), primary_key=True,
                  default=uuid.uuid4, unique=True, nullable=False)
    group_guid = Column(GUID(), ForeignKey('_ax_users.guid'))
    user_guid = Column(GUID(), ForeignKey('_ax_users.guid'))


class AxUser(Base):
    """Describes Ax users."""
    __tablename__ = '_ax_users'
    guid = Column(GUID(), primary_key=True,
                  default=uuid.uuid4, unique=True, nullable=False)
    name = Column(String(255))
    email = Column(String(255))
    password = Column(String(255))
    is_group = Column(Boolean, unique=False, default=False)
    is_admin = Column(Boolean, unique=False, default=False)
    is_everyone = Column(Boolean, unique=False, default=False)
    is_all_users = Column(Boolean, unique=False, default=False)
    is_anon = Column(Boolean, unique=False, default=False)
    parent = Column(GUID())
    position = Column(Integer)
    users = relationship(
        "AxUser",
        secondary='_ax_group2user',
        primaryjoin="AxUser.guid==AxGroup2Users.group_guid",
        secondaryjoin="AxUser.guid==AxGroup2Users.user_guid",
        backref="users_in_group"
    )


class AxRole2Users(Base):
    """What users and groups are assigned to Role?"""
    __tablename__ = '_ax_role2user'
    guid = Column(GUID(), primary_key=True,
                  default=uuid.uuid4, unique=True, nullable=False)
    role_guid = Column(GUID(), ForeignKey('_ax_roles.guid'))
    user_guid = Column(GUID(), ForeignKey('_ax_users.guid'))


class AxRole(Base):
    """Stores Roles for each workflow"""
    __tablename__ = '_ax_roles'
    guid = Column(GUID(), primary_key=True,
                  default=uuid.uuid4, unique=True, nullable=False)
    name = Column(String(255))
    form_guid = Column(GUID(), ForeignKey('_ax_forms.guid'))
    form = relationship("AxForm")
    users = relationship("AxUser", secondary='_ax_role2user')
    icon = Column(String(255)) # font-awesome key
    is_admin = Column(Boolean, unique=False, default=False)


class AxState2Role(Base):
    """Stores Roles that are assigned for state of workflow"""
    __tablename__ = '_ax_state2role'
    guid = Column(GUID(), primary_key=True,
                  default=uuid.uuid4, unique=True, nullable=False)
    state_guid = Column(GUID(), ForeignKey('_ax_states.guid'))
    role_guid = Column(GUID(), ForeignKey('_ax_roles.guid'))


class AxState(Base):
    """Stores states for each workflow"""
    __tablename__ = '_ax_states'
    guid = Column(GUID(), primary_key=True,
                  default=uuid.uuid4, unique=True, nullable=False)
    name = Column(String(255))
    form_guid = Column(GUID(), ForeignKey('_ax_forms.guid'))
    form = relationship("AxForm")
    roles = relationship("AxRole", secondary='_ax_state2role')
    is_start = Column(Boolean, unique=False, default=False)
    is_deleted = Column(Boolean, unique=False, default=False)
    is_all = Column(Boolean, unique=False, default=False)
    x = Column(Float)
    y = Column(Float)


class AxAction2Role(Base):
    """What roles are assigned for each action of workflow"""
    __tablename__ = '_ax_action2role'
    guid = Column(GUID(), primary_key=True,
                  default=uuid.uuid4, unique=True, nullable=False)
    action_guid = Column(GUID(), ForeignKey('_ax_actions.guid'))
    role_guid = Column(GUID(), ForeignKey('_ax_roles.guid'))


class AxAction(Base):
    """Stores actions for each workflow"""
    __tablename__ = '_ax_actions'
    guid = Column(GUID(), primary_key=True,
                  default=uuid.uuid4, unique=True, nullable=False)
    name = Column(String(255))
    db_name = Column(String(255))
    form_guid = Column(GUID(), ForeignKey('_ax_forms.guid'))
    form = relationship("AxForm")
    roles = relationship("AxRole", secondary='_ax_action2role')
    from_state_guid = Column(GUID(), ForeignKey('_ax_states.guid'))
    to_state_guid = Column(GUID(), ForeignKey('_ax_states.guid'))
    from_state = relationship('AxState', foreign_keys=[from_state_guid])
    to_state = relationship('AxState', foreign_keys=[to_state_guid])
    # Python code that is executed on this action. Sea action_schema.py
    code = Column(Text(convert_unicode=True))
    # If not None, the user will be see prompt before action is executed
    confirm_text = Column(String(255))
    # If True, the form will be closed after this action
    close_modal = Column(Boolean, unique=False, default=True)
    icon = Column(String(255)) # font-awesome key
    radius = Column(Float) # used in d3 worklfow constructor
    messages = None # Used to store messages and exceptions got from code
                    # Execution


class AxRoleFieldPermission(Base):
    """What role can view fields on each state of workflow"""
    __tablename__ = '_ax_role_field_permissions'
    guid = Column(GUID(), primary_key=True,
                  default=uuid.uuid4, unique=True, nullable=False)
    form_guid = Column(GUID(), ForeignKey('_ax_forms.guid'))
    form = relationship("AxForm")
    role_guid = Column(GUID(), ForeignKey('_ax_roles.guid'))
    role = relationship("AxRole")
    state_guid = Column(GUID(), ForeignKey('_ax_states.guid'))
    state = relationship("AxState")
    field_guid = Column(GUID(), ForeignKey('_ax_fields.guid'))
    field = relationship("AxField")
    read = Column(Boolean, unique=False, default=False)
    edit = Column(Boolean, unique=False, default=False)


class Ax1tomReference(Base):
    """Stores 1 to M fields relation"""
    __tablename__ = '_ax_mtom_references'
    guid = Column(GUID(), primary_key=True,
                  default=uuid.uuid4, unique=True, nullable=False)
    form_guid = Column(GUID(), ForeignKey('_ax_forms.guid'))
    field_guid = Column(GUID(), ForeignKey('_ax_fields.guid'))
    row_guid = Column(GUID())
    child_guid = Column(GUID())



""" This is part of GraphQL schema (Mutaions, Queryes, Subscriptions).
Workflow manipulation - create/update/delete - State, Action, Role,
Role2Action, Role2State, StatePermissions"""
import uuid
import graphene
# from loguru import logger
from backend.model import AxForm, AxField, AxAction, AxState, \
    AxRole, AxRoleFieldPermission, AxState2Role, AxAction2Role
import backend.model as ax_model
from backend.schemas.types import State, Action, \
    State2Role, Action2Role, RoleFieldPermission, Role


class CreateState(graphene.Mutation):
    """Create AxState

    Arguments:
        name (str): Name of created state. Usualy - default name from locale
        update (str): Default self-action name from locale. Like 'Update'
        form_guid (str): Guid of current AxForm
        x (float): X position of state in d3 workflow
        y (float): X position of state in d3 workflow

    Returns:
        state (AxState): Created state
        action (AxAction): Created update action
        permissions (List(AxRoleFieldPermission)): Default admin permissions
    """
    class Arguments:  # pylint: disable=missing-docstring
        name = graphene.String()
        update = graphene.String(required=False, default_value=None)
        form_guid = graphene.String()
        x = graphene.Float()
        y = graphene.Float()

    ok = graphene.Boolean()
    state = graphene.Field(State)
    action = graphene.Field(Action)
    permissions = graphene.List(RoleFieldPermission)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> CreateState"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            new_state = AxState()
            new_state.name = args.get('name')
            new_state.form_guid = args.get('form_guid')
            new_state.x = args.get('x')
            new_state.y = args.get('y')

            db_session.add(new_state)
            db_session.flush()

            # Create default self-action
            update_action = None
            if args.get('update'):
                update_action = AxAction()
                update_action.name = args.get('update')
                update_action.form_guid = uuid.UUID(args.get('form_guid'))
                update_action.from_state_guid = new_state.guid
                update_action.to_state_guid = new_state.guid
                db_session.add(update_action)

            # Add default admin role
            admin_role = db_session.query(AxRole).filter(
                AxRole.form_guid == uuid.UUID(args.get('form_guid'))
            ).filter(
                AxRole.is_admin.is_(True)
            ).first()

            # Add permissions for default admin role
            permissions = []
            if admin_role:
                state2role = AxState2Role()
                state2role.state_guid = new_state.guid
                state2role.role_guid = admin_role.guid
                db_session.add(state2role)

                ax_form = db_session.query(AxForm).filter(
                    AxForm.guid == uuid.UUID(args.get('form_guid'))
                ).first()

                for field in ax_form.fields:
                    perm = AxRoleFieldPermission()
                    perm.form_guid = ax_form.guid
                    perm.state_guid = new_state.guid
                    perm.role_guid = admin_role.guid
                    perm.field_guid = field.guid
                    perm.read = True
                    perm.edit = True
                    db_session.add(perm)
                    permissions.append(perm)

            return CreateState(
                state=new_state,
                action=update_action,
                permissions=permissions,
                ok=True)


class UpdateState(graphene.Mutation):
    """Update AxState"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        name = graphene.String()
        x = graphene.Float()
        y = graphene.Float()

    ok = graphene.Boolean()
    state = graphene.Field(State)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> UpdateState"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            ax_state = db_session.query(AxState).filter(
                AxState.guid == uuid.UUID(args.get('guid'))
            ).first()
            if args.get('name'):
                ax_state.name = args.get('name')
            if args.get('x'):
                ax_state.x = args.get('x')
            if args.get('y'):
                ax_state.y = args.get('y')
            return UpdateState(state=ax_state, ok=True)


class DeleteState(graphene.Mutation):
    """Deletes AxState"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> DeleteState"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            ax_state = db_session.query(AxState).filter(
                AxState.guid == uuid.UUID(guid)
            ).first()

            db_session.query(AxAction).filter(
                AxAction.from_state_guid == uuid.UUID(guid)
            ).delete()

            db_session.query(AxAction).filter(
                AxAction.to_state_guid == uuid.UUID(guid)
            ).delete()

            db_session.delete(ax_state)
            return DeleteState(deleted=guid, ok=True)


class CreateAction(graphene.Mutation):
    """Creates AxAction"""
    class Arguments:  # pylint: disable=missing-docstring
        form_guid = graphene.String()
        name = graphene.String()
        from_state_guid = graphene.String()
        to_state_guid = graphene.String()

    ok = graphene.Boolean()
    action = graphene.Field(Action)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> CreateAction"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            new_action = AxAction()
            new_action.name = args.get('name')
            new_action.form_guid = args.get('form_guid')
            new_action.from_state_guid = args.get('from_state_guid')
            new_action.to_state_guid = args.get('to_state_guid')
            db_session.add(new_action)

            # Add default admin role
            admin_role = db_session.query(AxRole).filter(
                AxRole.form_guid == uuid.UUID(args.get('form_guid'))
            ).filter(
                AxRole.is_admin.is_(True)
            ).first()

            if admin_role:
                action2role = AxAction2Role()
                action2role.action_guid = new_action.guid
                action2role.role_guid = admin_role.guid
                db_session.add(action2role)

            return CreateAction(action=new_action, ok=True)


class UpdateAction(graphene.Mutation):
    """Updates AxAction"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        name = graphene.String(required=False, default_value=None)
        db_name = graphene.String(required=False, default_value=None)
        code = graphene.String(required=False, default_value=None)
        confirm_text = graphene.String(required=False, default_value=None)
        close_modal = graphene.Boolean(required=False, default_value=None)
        icon = graphene.String(required=False, default_value=None)
        radius = graphene.Float(required=False, default_value=None)

    ok = graphene.Boolean()
    action = graphene.Field(Action)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> UpdateAction"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            ax_action = db_session.query(AxAction).filter(
                AxAction.guid == uuid.UUID(args.get('guid'))
            ).first()

            if args.get('name'):
                ax_action.name = args.get('name')
            if args.get('db_name'):
                ax_action.db_name = args.get('db_name')
            if args.get('code'):
                ax_action.code = args.get('code')
            if args.get('confirm_text'):
                ax_action.confirm_text = args.get('confirm_text')
            if args.get('close_modal') is not None:
                ax_action.close_modal = args.get('close_modal')
            if args.get('icon'):
                ax_action.icon = args.get('icon')
            if args.get('radius') is not None:
                ax_action.radius = args.get('radius')

            return UpdateAction(action=ax_action, ok=True)


class DeleteAction(graphene.Mutation):
    """Deletes AxAction"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> DeleteAction"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            ax_action = db_session.query(AxAction).filter(
                AxAction.guid == uuid.UUID(guid)
            ).first()
            db_session.delete(ax_action)
            return DeleteAction(deleted=guid, ok=True)


class CreateRole(graphene.Mutation):
    """Creates AxRole"""
    class Arguments:  # pylint: disable=missing-docstring
        form_guid = graphene.String()
        name = graphene.String()

    ok = graphene.Boolean()
    role = graphene.Field(Role)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> CreateRole"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            ax_role = AxRole()
            ax_role.name = args.get('name')
            ax_role.form_guid = args.get('form_guid')
            db_session.add(ax_role)
            return CreateRole(role=ax_role, ok=True)


class UpdateRole(graphene.Mutation):
    """Updates AxRole"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        name = graphene.String(required=False, default_value=None)
        icon = graphene.String(required=False, default_value=None)

    ok = graphene.Boolean()
    role = graphene.Field(Role)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> UpdateRole"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            ax_role = db_session.query(AxRole).filter(
                AxRole.guid == uuid.UUID(args.get('guid'))
            ).first()
            if args.get('name'):
                ax_role.name = args.get('name')
            if args.get('icon'):
                ax_role.icon = args.get('icon')
            return UpdateRole(role=ax_role, ok=True)


class DeleteRole(graphene.Mutation):
    """Deletes AxRole"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> DeleteRole"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            ax_role = db_session.query(AxRole).filter(
                AxRole.guid == uuid.UUID(guid)
            ).first()

            db_session.delete(ax_role)
            return DeleteRole(deleted=guid, ok=True)


class AddRoleToState(graphene.Mutation):
    """Create AxState2Role"""
    class Arguments:  # pylint: disable=missing-docstring
        state_guid = graphene.String()
        role_guid = graphene.String()

    ok = graphene.Boolean()
    state2role = graphene.Field(State2Role)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> AddRoleToState"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            role_guid = args.get('role_guid')
            state_guid = args.get('state_guid')
            existing_state2role = db_session.query(
                AxState2Role
            ).filter(
                AxState2Role.role_guid == uuid.UUID(role_guid)
            ).filter(
                AxState2Role.state_guid == uuid.UUID(state_guid)
            ).first()

            if existing_state2role:
                ok = True
                return AddRoleToState(state2role=existing_state2role, ok=ok)

            state2role = AxState2Role()
            state2role.state_guid = uuid.UUID(state_guid)
            state2role.role_guid = uuid.UUID(role_guid)

            db_session.add(state2role)
            return AddRoleToState(state2role=state2role, ok=True)


class DeleteRoleFromState(graphene.Mutation):
    """Deletes AxState2Role"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        role_guid = graphene.String()
        state_guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()
    role_guid = graphene.String()
    state_guid = graphene.String()

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> DeleteRoleFromState"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            role_guid = args.get('role_guid')
            state_guid = args.get('state_guid')

            role2state = None
            if guid:
                role2state = db_session.query(AxState2Role).filter(
                    AxState2Role.guid == uuid.UUID(guid)
                ).first()
            else:
                role2state = db_session.query(AxState2Role).filter(
                    AxState2Role.role_guid == uuid.UUID(role_guid)
                ).filter(
                    AxState2Role.state_guid == uuid.UUID(state_guid)
                ).first()

            the_role_guid = role2state.role_guid
            the_state_guid = role2state.state_guid

            db_session.delete(role2state)
            return DeleteRoleFromState(
                deleted=guid,
                role_guid=the_role_guid,
                state_guid=the_state_guid,
                ok=True)


class AddRoleToAction(graphene.Mutation):
    """Create AxAction2Role"""
    class Arguments:  # pylint: disable=missing-docstring
        action_guid = graphene.String()
        role_guid = graphene.String()

    ok = graphene.Boolean()
    action2role = graphene.Field(Action2Role)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> AddRoleToAction"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            role_guid = args.get('role_guid')
            action_guid = args.get('action_guid')
            existing_action2role = db_session.query(
                AxAction2Role
            ).filter(
                AxAction2Role.role_guid == uuid.UUID(role_guid)
            ).filter(
                AxAction2Role.action_guid == uuid.UUID(action_guid)
            ).first()

            if existing_action2role:
                ok = True
                return AddRoleToAction(action2role=existing_action2role, ok=ok)

            action2role = AxAction2Role()
            action2role.action_guid = uuid.UUID(action_guid)
            action2role.role_guid = uuid.UUID(role_guid)

            db_session.add(action2role)
            return AddRoleToAction(action2role=action2role, ok=True)


class DeleteRoleFromAction(graphene.Mutation):
    """Deletes AxAction2Role"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        action_guid = graphene.String()
        role_guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()
    role_guid = graphene.String()
    action_guid = graphene.String()

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> DeleteRoleFromAction"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            action_guid = args.get('action_guid')
            role_guid = args.get('role_guid')

            role2action = None
            if guid:
                role2action = db_session.query(AxAction2Role).filter(
                    AxAction2Role.guid == uuid.UUID(guid)
                ).first()
            else:
                role2action = db_session.query(AxAction2Role).filter(
                    AxAction2Role.role_guid == uuid.UUID(role_guid)
                ).filter(
                    AxAction2Role.action_guid == uuid.UUID(action_guid)
                ).first()

            deleted = role2action.guid
            db_session.delete(role2action)
            return DeleteRoleFromAction(
                deleted=deleted,
                ok=True,
                role_guid=role_guid,
                action_guid=action_guid)


class SetStatePermission(graphene.Mutation):
    """Creates AxRoleFieldPermission or multiple if tab field is presented"""
    class Arguments:  # pylint: disable=missing-docstring
        form_guid = graphene.String()
        state_guid = graphene.String()
        role_guid = graphene.String()
        field_guid = graphene.String()
        read = graphene.Boolean()
        edit = graphene.Boolean()

    ok = graphene.Boolean()
    permissions = graphene.List(RoleFieldPermission)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "workflow_schema -> SetStatePermission"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            form_guid = args.get('form_guid')
            state_guid = args.get('state_guid')
            role_guid = args.get('role_guid')
            field_guid = args.get('field_guid')
            read = args.get('read')
            edit = args.get('edit')

            current_fields_guids = []

            ax_field = db_session.query(AxField).filter(
                AxField.guid == uuid.UUID(field_guid)).first()

            if not ax_field.is_tab:
                current_fields_guids.append(ax_field.guid)
            else:
                for field in ax_field.form.fields:
                    if field.parent == ax_field.guid:
                        current_fields_guids.append(field.guid)

            for current_guid in current_fields_guids:
                ax_perm = db_session.query(
                    AxRoleFieldPermission
                ).filter(
                    AxRoleFieldPermission.state_guid == uuid.UUID(state_guid)
                ).filter(
                    AxRoleFieldPermission.role_guid == uuid.UUID(role_guid)
                ).filter(
                    AxRoleFieldPermission.field_guid == current_guid
                ).first()

                if ax_perm is None:
                    ax_perm = AxRoleFieldPermission()
                    ax_perm.form_guid = uuid.UUID(form_guid)
                    ax_perm.role_guid = uuid.UUID(role_guid)
                    ax_perm.field_guid = current_guid
                    ax_perm.state_guid = uuid.UUID(state_guid)
                    ax_perm.read = read
                    ax_perm.edit = edit
                    db_session.add(ax_perm)
                else:
                    ax_perm.read = read
                    ax_perm.edit = edit

                db_session.flush()

            return_permissions = db_session.query(
                AxRoleFieldPermission
            ).filter(
                AxRoleFieldPermission.state_guid == uuid.UUID(state_guid)
            ).all()

            return SetStatePermission(permissions=return_permissions, ok=True)


class WorkflowMutations(graphene.ObjectType):
    """Combine all mutations"""
    create_state = CreateState.Field()
    update_state = UpdateState.Field()
    delete_state = DeleteState.Field()
    create_action = CreateAction.Field()
    update_action = UpdateAction.Field()
    delete_action = DeleteAction.Field()
    create_role = CreateRole.Field()
    update_role = UpdateRole.Field()
    delete_role = DeleteRole.Field()
    add_role_to_state = AddRoleToState.Field()
    delete_role_from_state = DeleteRoleFromState.Field()
    add_role_to_action = AddRoleToAction.Field()
    delete_role_from_action = DeleteRoleFromAction.Field()
    set_state_permission = SetStatePermission.Field()



"""AxNum field type functions - before, after / insert, update, delete"""
import sys
import traceback
import ujson as json
from backend.model import AxMetric


async def before_insert(db_session, field, before_form, tobe_form, action,
                        current_user):
    """
    Executes python code from AxField.private_options

    Returns:
        Object: Returns updated value of current field"""
    del before_form, action, current_user, tobe_form

    if not field.private_options_json or field.private_options_json == '{}':
        return None

    # get current counter from AxMetric for current key
    # if there is none -> create one

    options = json.loads(field.private_options_json)
    code = options['algorithm']
    key = options['counterKey']

    if not key:
        return None

    current_counter = db_session.query(AxMetric).filter(
        AxMetric.key == key
    ).first()

    current_value = None
    if current_counter:
        current_value = current_counter.value
    else:
        current_counter = AxMetric()
        current_counter.key = key
        current_counter.value = None
        db_session.add(current_counter)

    localz = dict()
    localz['ax_counter'] = current_value
    localz['ax_num'] = current_value

    try:
        exec(code, globals(), localz)  # pylint: disable=W0122
        current_counter.value = localz['ax_counter']
        field.needs_sql_update = True
        field.value = localz['ax_num']
    except SyntaxError as err:
        error_class = err.__class__.__name__
        detail = err.args[0]
        line_number = err.lineno
        ret_data = {
            "exception": {
                "error_class": error_class,
                "line_number": line_number,
                "detail": detail
            },
        }
        field.value = json.dumps(ret_data)
    except Exception as err:  # pylint: disable=W0703
        error_class = err.__class__.__name__
        detail = err.args[0]
        cl, exc, tb = sys.exc_info()
        del cl, exc
        line_number = traceback.extract_tb(tb)[-1][1]
        del tb
        ret_data = {
            "exception": {
                "error_class": error_class,
                "line_number": line_number,
                "action_name": action.name,
                "detail": detail
            }
        }
        field.value = json.dumps(ret_data)

    return field.value



"""All routes of ax-core and ax-admin"""

import os
import sys
import asyncio
import uuid

from sanic import response
from loguru import logger
import aiopubsub
from graphql_ws.websockets_lib import WsLibSubscriptionServer
from sanic_graphql import GraphQLView
from graphql.execution.executors.asyncio import AsyncioExecutor
import ujson as json

import backend.cache as ax_cache
import backend.schema as ax_schema
import backend.misc as ax_misc
import backend.pubsub as ax_pubsub
from backend.tus import tus_bp
import backend.model as ax_model
from backend.model import AxForm, AxField
import backend.schemas.form_schema as form_schema
# import backend.scheduler as ax_scheduler
import backend.dialects as ax_dialects

this = sys.modules[__name__]

loop = asyncio.new_event_loop()
actions_loop = asyncio.new_event_loop()
app = None
graphql_view = None
dummy_view = None


class AxGraphQLView(GraphQLView):
    """ Extends GraphQLView to output GQL errors and schema updates"""

    def __init__(self, **kwargs):
        db_session = ax_model.Session()
        super().__init__(
            schema=ax_schema.schema,
            graphiql=False,
            enable_async=True,
            executor=AsyncioExecutor(loop=this.loop),
            context={'session': db_session}
        )

    @staticmethod
    def format_error(error):
        """ This method is required for showing graphql errors """
        logger.error(error)
        return GraphQLView.format_error(error)


def init_graphql_view():  # pylint: disable=unused-variable
    """Initiate graphql"""
    this.graphql_view = AxGraphQLView.as_view()
    this.app.add_route(this.graphql_view, '/api/graphql')


# async def do_action():
#     """ Test function """
#     from backend.schema import schema

#     values = {
#         "searchString": "Chateau Lafite Rothschild 1996",
#         "supplier": "9c198568469447268d62b395badeb71f",
#         "loader": "0eb38c45e9954442995c6205f2e3f7eb"
#     }

#     query = f"""
#         mutation(
#             $formDbName: String,
#             $actionDbName: String,
#             $values: String
#         ){{
#             doAction(
#                 formDbName: $formDbName
#                 actionDbName: $actionDbName
#                 values: $values
#             ) {{
#                 form {{
#                 guid
#                 dbName
#                 }}
#                 newGuid
#                 messages
#                 ok
#             }}
#         }}
#     """
#     variables = {
#         "formDbName": "Stock",
#         "actionDbName": "addDraft",
#         "values": json.dumps(values)
#     }
#     result = schema.execute(query, variables=variables)
#     test_str = json.dumps(result.data, sort_keys=True, indent=4)
#     print(test_str)


def init_routes(sanic_app, deck_path=None):
    """Innitiate all Ax routes"""
    try:

        sanic_app.static('/uploads', str(ax_misc.path('uploads')))
        sanic_app.static('/static', str(ax_misc.path('dist/ax/static')))
        sanic_app.static('/stats', str(ax_misc.path('dist/ax/stats.html')))
        sanic_app.static(
            '/test_webpack', str(ax_misc.path('dist/ax/test.html')))

        deck_dist_path = str(ax_misc.path('dist/deck'))
        if deck_path:
            deck_dist_path = deck_path
        index_path = str(os.path.join(deck_dist_path, "index.html"))

        sanic_app.static('/deck/', deck_dist_path)
        sanic_app.static('/deck', index_path)

        # Add tus upload blueprint
        sanic_app.blueprint(tus_bp)
        # Add web-socket subscription server
        subscription_server = WsLibSubscriptionServer(ax_schema.schema)

        @sanic_app.route(
            '/api/file/<form_guid>/<row_guid>/<field_guid>/<file_name>',
            methods=['GET'])
        async def db_file_viewer(   # pylint: disable=unused-variable
                request, form_guid, row_guid, field_guid, file_name):  # pylint: disable=unused-variable
            """ Used to display files that are stored in database.
                Used in fields like AxImageCropDb"""
            del request, form_guid, file_name
            with ax_model.scoped_session("routes.db_file_viewer") as db_session:
                safe_row_guid = str(uuid.UUID(str(row_guid)))
                ax_field = db_session.query(AxField).filter(
                    AxField.guid == uuid.UUID(field_guid)
                ).first()
                field_value = await ax_dialects.dialect.select_field(
                    db_session=db_session,
                    form_db_name=ax_field.form.db_name,
                    field_db_name=ax_field.db_name,
                    row_guid=safe_row_guid)
                return response.raw(
                    field_value, content_type='application/octet-stream')

        @sanic_app.route(
            '/api/file/<form_guid>/<row_guid>/<field_guid>/<file_guid>/<file_name>',    # pylint: disable=line-too-long
            methods=['GET'])
        async def file_viewer(  # pylint: disable=unused-variable
                request, form_guid, row_guid, field_guid, file_guid, file_name):
            """ Used to display files uploaded and stored on disk.
                Displays temp files too. Used in all fields with upload"""
            del request
            with ax_model.scoped_session("routes -> file_viewer") as db_session:
                # if row_guid is null -> display from /tmp without permissions
                if not row_guid or row_guid == 'null':
                    tmp_dir = os.path.join(ax_misc.tmp_root_dir, file_guid)
                    file_name = os.listdir(tmp_dir)[0]
                    temp_path = os.path.join(tmp_dir, file_name)
                    return await response.file(temp_path)

                # get AxForm with row values
                ax_form = db_session.query(AxForm).filter(
                    AxForm.guid == uuid.UUID(form_guid)
                ).first()
                ax_form = await form_schema.set_form_values(
                    db_session=db_session,
                    ax_form=ax_form,
                    row_guid=row_guid)

                # Get values from row, field
                field_values = None
                for field in ax_form.fields:
                    if field.guid == uuid.UUID(field_guid):
                        if field.value:
                            field_values = json.loads(field.value)

                # Find requested file in value
                the_file = None
                for file in field_values:
                    if file['guid'] == file_guid:
                        the_file = file

                if not the_file:
                    return response.text("", status=404)

                # if file exists -> return file
                row_guid_str = str(uuid.UUID(row_guid))
                file_path = os.path.join(
                    ax_misc.uploads_root_dir,
                    'form_row_field_file',
                    form_guid,
                    row_guid_str,
                    field_guid,
                    the_file['guid'],
                    the_file['name'])
                if not os.path.lexists(file_path):
                    return response.text("", status=404)
                return await response.file(file_path)

        @sanic_app.route('/admin/<path:path>')
        @sanic_app.route('/form/<path:path>')
        def index(request, path):  # pylint: disable=unused-variable
            """ This is MAIN ROUTE. (except other routes listed in this module).
                All requests are directed to Vue single page app. After that Vue
                handles routing."""
            del request, path
            absolute_path = ax_misc.path('dist/ax/index.html')
            return response.html(open(absolute_path).read())

        @sanic_app.route('/')
        def handle_request(request):  # pylint: disable=unused-variable
            del request
            return response.redirect('/deck')

        @sanic_app.route('/draw_ax')
        async def draw_ax(request):  # pylint: disable=unused-variable
            """ Outputs bundle.js. Used when Ax web-components
                are inputed somewhere. Users can use this url for <script> tag
                """
            del request
            absolute_path = ax_misc.path('dist/ax/static/js/ax-bundle.js')
            return await response.file(
                absolute_path,
                headers={
                    'Content-Type': 'application/javascript; charset=utf-8'
                }
            )

        @sanic_app.websocket('/api/subscriptions', subprotocols=['graphql-ws'])
        async def subscriptions(request, web_socket):  # pylint: disable=unused-variable
            """Web socket route for graphql subscriptions"""
            del request
            try:
                # TODO: Why socket error exception occurs without internet
                await subscription_server.handle(web_socket)
                return web_socket
            except asyncio.CancelledError:
                pass
                # logger.exception('Socket error')

        @sanic_app.route('/api/ping')
        async def ping(request):
            """ Ping function checks if Ax is running. Used with monit """
            from backend.schema import schema
            result = schema.execute("query { ping }")
            test_str = json.dumps(result.data, sort_keys=True, indent=4)
            return response.text(test_str)

        @sanic_app.route('/api/test')
        async def test(request):  # pylint: disable=unused-variable
            """Test function"""
            del request
            this.test_schema = 'IT WORKS'
            ax_pubsub.publisher.publish(
                aiopubsub.Key('dummy_test'), this.test_schema)
            return response.text(this.test_schema)

        @sanic_app.route('/api/set')
        async def cache_set(request):  # pylint: disable=unused-variable
            """Cache Test function"""
            del request
            obj = ['one', 'two', 'three']
            await ax_cache.cache.set('user_list', obj)
            return response.text('Cache SET' + str(obj))

        @sanic_app.route('/api/get')
        async def cache_get(request):  # pylint: disable=unused-variable
            """Cache Test function"""
            del request
            obj = await ax_cache.cache.get('user_list')
            ret_str = 'READ cache == ' + \
                str(obj[0].username + ' - ' + os.environ['AX_VERSION'])
            return response.text(ret_str)

    except Exception:
        logger.exception('Error initiating routes.')
        raise



"""Miscellaneous functions
"""

import os
import sys
import uuid
from pathlib import Path
from datetime import datetime
import graphene
import graphene_sqlalchemy
import pytz
import yaml
from loguru import logger

this = sys.modules[__name__]
root_path = None
timezone = None
tmp_root_dir = None
uploads_root_dir = None


def init_misc(timezone_name, tmp_absolute_path, uploads_absolute_path) -> None:
    """Inititate time_zone, root_path"""

    try:
        this.uploads_root_dir = path('uploads')
        if uploads_absolute_path is not None:
            this.uploads_root_dir = str(Path(uploads_absolute_path))
        if not server_is_app_engine():
            if not os.path.exists(this.uploads_root_dir):
                os.makedirs(this.uploads_root_dir)
    except Exception:
        logger.exception(
            'Uploads folder does not exist and Ax cant create it')
        raise

    try:
        this.tmp_root_dir = path('tmp')
        if tmp_absolute_path is not None:
            this.tmp_root_dir = str(Path(tmp_absolute_path))
        if not os.path.exists(this.tmp_root_dir):
            if not server_is_app_engine():
                os.makedirs(this.tmp_root_dir)
    except Exception:
        logger.exception(
            'Tmp folder does not exist and Ax cant create it')
        raise

    if timezone_name not in pytz.common_timezones:
        logger.error(
            'Timezone {tz} not valid. Falling back to UTC.', tz=timezone_name)
        this.timezone = 'UTC'

    this.timezone = pytz.timezone(timezone_name)


def date(_date: datetime = datetime.now()) -> datetime:
    """Localise python datetime"""
    localized_date = this.timezone.localize(_date, is_dst=None)
    return localized_date


def path(_path: str = '') -> str:
    """Get correct absolute path.

    Agrs:
        _path (str, optional): Defaults to ''. Path relative to module root.

    Returns:
        str: Absolute path to specified file or directory
    Examples:
        Access module files like path('backend/schemas/user_schema.py')
    """
    if this.root_path is None:
        this.root_path = Path(__file__).resolve().parent.parent
        # logger.debug('File path = {root}', root=Path(__file__))
    return root_path / _path


def server_is_app_engine() -> bool:
    """Is Sanic is running on Google App Engine Standard"""
    return os.getenv("GAE_VERSION") or None


def load_configuration() -> None:
    """Function that loads ax application configuration from app.yaml

    If application is running on Google App Engine, all params from
    env_variables section of app.yaml will be loaded as enviroment variables.
    If application is running outside of App Engine then we must load
    variables ourselves.

    Args:
        app_yaml_location (str): Path to app.yaml location.

    Returns:
        bool: The return value. True for success.
    Examples:
        Access configuration variables like os.environ['AX_VERSION']

    Raises:
        FileNotFoundError: Cant find app.yaml
        ValueError: Cant parse app.yaml
        LookupError: No env_variables section in app.yaml
    """
    app_yaml = path('app.yaml')

    if not os.path.isfile(app_yaml):
        raise FileNotFoundError('Configuration failed, app.yaml not found')
    if not server_is_app_engine():
        with open(app_yaml, 'r') as stream:
            try:
                yaml_vars = yaml.safe_load(stream)
                if 'env_variables' not in yaml_vars:
                    err = 'Configuration failed, no env_variables in app.yaml'
                    logger.error(err)
                    raise LookupError(err)
                for key, value in yaml_vars['env_variables'].items():
                    if value:
                        os.environ[key] = str(value)
            except yaml.YAMLError as exc:
                err = 'Configuration failed, cant parse yaml - ' + exc
                logger.error(err)
                raise ValueError(err)


def convert_column_to_string(type, column, registry=None):
    """This functions is needed to convert UUID column to String column
    for SQL Alchemy"""
    return graphene.String


def string_is_guid(string_guid):
    """Checks if string is valid guid can be
     302a4299736e4ef384fca9f400e84b24 or 302a4299-736e-4ef3-84fc-a9f400e84b24

    Returns:
        [str]: version of uuid or None
    """
    try:
        return uuid.UUID(string_guid).version
    except ValueError:
        return None



"""Scheduler using APScheduller
1 - Test scheduler Job - 5 sec after sanic start
2 - Clean /tmp folder - every 30 mins
"""
import os
import sys
import time
import shutil
import asyncio
from datetime import timedelta, datetime
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.executors.asyncio import AsyncIOExecutor
from apscheduler.jobstores.sqlalchemy import SQLAlchemyJobStore
# from graphql.execution.executors.asyncio import AsyncioExecutor
from loguru import logger
# import ujson as json
import backend.model as ax_model
import backend.misc as ax_misc

this = sys.modules[__name__]
scheduler = None


async def prn_job(message):
    """ Executed aftrer sanic app start to know if scheduler is working"""
    print(str(message))
    return False


async def clear_tmp_files():
    """ Delete all files from /tmp folder wich is expired """
    tmp_folder = ax_misc.tmp_root_dir
    for root, dirs, _ in os.walk(tmp_folder):
        del root
        for dir_name in dirs:
            dir_to_check = os.path.join(tmp_folder, dir_name)
            seconds = time.time() - os.path.getmtime(dir_to_check)
            minutes = int(seconds) / 60
            if minutes > 120:   # 120 minutes expiration time
                shutil.rmtree(dir_to_check)


async def gql_query_job(query, variables):
    """ Jon that executes Graphql query """
    from backend.schema import schema
    # result = schema.execute(query, variables=variables)

    await schema.execute(
        query,
        variables=variables,
        return_promise=True)
#         executor=AsyncioExecutor(loop=asyncio.get_event_loop()),


# def do_action_REM(form_db_name, action_db_name, values, row_guid=None):
#     """
#     Adds job that executes GQL query for workflow action. Used in AxAction
#     python code.

#     Args:
#         form_db_name (str): db_name of AxForm
#         action_db_name (str): db_name of AxAction that needs to be performed
#         values (Dict): Dict with form values. Keys are db table columns
#         row_guid (str): Guid (not dashes) of database row

#     """

#     query = f"""
#         mutation(
#             $formDbName: String,
#             $actionDbName: String,
#             $rowGuid: String,
#             $values: String
#         ){{
#             doAction(
#                 formDbName: $formDbName
#                 actionDbName: $actionDbName
#                 rowGuid: $rowGuid
#                 values: $values
#             ) {{
#                 form {{
#                 guid
#                 dbName
#                 }}
#                 newGuid
#                 messages
#                 ok
#             }}
#         }}
#     """
#     variables = {
#         "formDbName": form_db_name,
#         "actionDbName": action_db_name,
#         "rowGuid": row_guid,
#         "values": json.dumps(values)
#     }
#     args = {
#         "query": query,
#         "variables": variables
#     }
#     job = this.scheduler.add_job(func=gql_query_job, kwargs=args,
#                                  misfire_grace_time=60 * 60)
#     print(str(job))


def add_action_job(form_db_name, action_db_name, values, row_guid=None,
                   wait_seconds=0):
    """ Adds scheduller job that executes a AxAction """
    import backend.schemas.action_schema as ax_action_schema
    variables = {
        "form_db_name": form_db_name,
        "action_db_name": action_db_name,
        "row_guid": row_guid,
        "values": values
    }
    moscow_dt = datetime.now() + timedelta(seconds=wait_seconds)

    job = this.scheduler.add_job(
        func=ax_action_schema.execute_action,
        kwargs=variables,
        misfire_grace_time=60 * 60,
        trigger='date',
        next_run_time=moscow_dt)
    print(str(job))


def init_scheduler():
    """Initiate scheduller"""
    # https://apscheduler.readthedocs.io/en/latest/modules/schedulers/base.html#apscheduler.schedulers.base.BaseScheduler.add_job

    # TODO UTC must match timezone from config
    try:
        jobstores = {
            'default': SQLAlchemyJobStore(
                engine=ax_model.engine,
                tablename='_ax_scheduler_jobs'
            )
        }

        executors = {
            'default': AsyncIOExecutor()
        }

        loop = asyncio.get_event_loop()
        this.scheduler = AsyncIOScheduler(
            jobstores=jobstores,
            executors=executors,
            timezone=ax_misc.timezone,
            event_loop=loop)
        this.scheduler.start()

        # this.scheduler.remove_all_jobs()
        # dt = datetime.now() + timedelta(seconds=5)

        moscow_dt = ax_misc.date() + timedelta(seconds=5)
        this.scheduler.add_job(
            prn_job,
            'date',
            run_date=moscow_dt,
            args=['SCHEDULER WORKS'],
            id='prn_job',
            misfire_grace_time=60
        )

        # Job cleaning /uploads/tmp folder. deletes files that are expired
        all_jobs = this.scheduler.get_jobs()
        if "clear_tmp_files" not in [job.name for job in all_jobs]:
            this.scheduler.add_job(
                clear_tmp_files,
                trigger='cron',
                minute='30',
                coalesce=True,
                misfire_grace_time=60,
                id='clear_tmp_files')

    except Exception:
        logger.exception('Error initiating scheduler module. ')
        raise
